#!/usr/bin/env python

import optparse
import random
import sys
import traceback
import os
import multiprocessing
from glob import glob
from datetime import datetime

import exp
import util
import gedf
import model.serialize as s


from util.stats import stats
from overheads import const, CPMD

o = optparse.make_option
opts = [
    o('-s', '--scheduler', action='store', dest='sched',
      help='which scheduler to use'),

    o('-m', '--machine', action='store', dest='machine',
      help='which machine config to use'),

    o('-q', '--quanta', action='store', dest='quanta',
      type='choice', choices=['staggered', 'aligned']),

    o('-p', '--prefix', action='store', dest='prefix',
      help='where to store result files (default: STDOUT)'),

    o('-r', '--random', action='store_true', dest='randorder',
      help='process files in random order'),

    o('-g', '--glob', action='store_true', dest='glob',
      help='interpret argument as pattern'),

    o(None, '--parallel', action='store_true', dest='parallel',
      help='use all available cores'),

    o(None, '--overheads', action='store', dest='ohead_dir',
      help='where to load overhead data'),

    o(None, '--no-header', action='store_false', dest='header',
      help='suppress header line'),

    o(None, '--dump-header', action='store_true', dest='dump_header',
      help='print header and quit'),
    ]

options  = None
defaults = {
    # Runtime config.
    'prefix'    : None,
    'ohead_dir' : util.get_this_repo() + '/ohead-data',

    'randorder' : False,
    'parallel'  : False,
    'glob'      : False,
    'progress'  : 0,
    'verbose'    : False,

    'header'    : True,
    'dump_header' : False,

    # Experiment setup.
    'machine'    : 'ludwig',
    'sched'      : 'P-EDF',
    'quanta'     :  None,
    }

VERSION  = util.git_current_head()
MODIFIED = util.git_locally_modified()

HOSTS = {
    'ludwig'  : 24,
}

def log(msg):
    print "[%s] %s" % (datetime.now(), msg)

def verbose(msg):
    if options.verbose:
        log(msg)

def get_test(wss, hrt, background):
    host = options.machine
    cpus = HOSTS[host]

    (sched_test, cache_level, dedicated_irq, is_global) = \
        exp.diss.get_sched_test(host, options.sched, hrt)

    if is_global and dedicated_irq:
        cpus -= 1 # one is reserved

    (oheads, cpmd) = exp.diss.get_overheads(
        host, options.sched, background, hrt,
        quanta=options.quanta)

    idx = CPMD.get_idx_for_name(cache_level)
    oheads.cache_affinity_loss = const(cpmd.mem_hierarchy[idx](wss))
    oheads.initial_cache_load  = const(cpmd.MEM(wss))

    kargs = {}
    kargs['dedicated_irq'] = dedicated_irq
    kargs['total_cpus']    = HOSTS[host]

    if options.quanta:
        kargs['staggered'] = options.quanta == 'staggered'

    def run_test(taskset):
        return sched_test(cpus, oheads, taskset, **kargs)

    return run_test

def run_hrt_test(tp, wss, bg):
    sched_test = get_test(wss, True, bg)
    total = len(tp.tasksets)
    ok = 0
    for i, ts in enumerate(tp.tasksets):
        if options.progress and (i+1) % options.progress == 0:
            log("HRT-%s taskset %d of %d..." % (bg, (i+1), total))
        if sched_test(ts.copy()):
            ok += 1
    return [ok / float(total)]

def tardiness(t):
    return max(0, t.completion - t.deadline)

def rel_tardiness(t):
    return tardiness(t) / float(t.deadline)

def tardiness_stats(ts):
    tard  = [tardiness(t) for t in ts]
    return [x for x in stats(tard, want_max=True)]

def rel_tardiness_stats(ts):
    tard  = [rel_tardiness(t) for t in ts]
    return [x for x in stats(tard, want_max=True)]

def run_srt_test(tp, wss, bg):
    sched_test = get_test(wss, False, bg)
    total = len(tp.tasksets)
    ok = 0
    collected = []
    for i, ts in enumerate(tp.tasksets):
        if options.progress and (i+1) % options.progress == 0:
            log("SRT-%s taskset %d of %d..." % (bg, (i+1), total))
        res = sched_test(ts.copy())
        if res:
            ok += 1
            collected.extend(res)

    return [ok / float(total)] + \
        tardiness_stats(collected) + \
        rel_tardiness_stats(collected)

def get_outfile(fname):
    if not options.prefix is None:
        output = options.prefix + '/' + fname
        return util.io.atomic_create_file(output)
    else:
        return sys.stdout

def target_file(fname):
    bname = os.path.basename(fname)
    name, ext = os.path.splitext(bname)
    if options.quanta:
        qstr = '_quanta=%s' % options.quanta
    else:
        qstr = ''
    return "%s_host=%s_scheduler=%s%s.csv" % \
        (name, options.machine, options.sched, qstr)

HEADERS = [
    'Distribution',
    'Deadlines',
    'Host',
    'Scheduler',
    'WSS',
    'UCap',
    'HRT (idle)',
    'HRT (load)',
    'SRT (idle)',
    'T-MAX (idle)',
    'T-AVG (idle)',
    'T-MED (idle)',
    'T-STD (idle)',
    'RT-MAX (idle)',
    'RT-AVG (idle)',
    'RT-MED (idle)',
    'RT-STD (idle)',
    'SRT (load)',
    'T-MAX (load)',
    'T-AVG (load)',
    'T-MED (load)',
    'T-STD (load)',
    'RT-MAX (load)',
    'RT-AVG (load)',
    'RT-MED (load)',
    'RT-STD (load)',
]

def print_header(out):
    util.print_rows([HEADERS], col_sep=', ', col_width=20,
                    f=out, prepend='#')

def process_file(fname):
    tname = target_file(fname)
    out = get_outfile(tname)
    if out:
        tp = s.load(fname)
        # load
        if tp is None or tp.xml.tag != 'testpoint':
            print "FAILED: couldn't load testpoint data."
            return

        dist = tp.params['dist']
        dl   = tp.params['deadlines']
        ucap = float(tp.params['ucap'])
        wss  = int(tp.params['wss'])

        verbose('Config: dist=%s deadlines=%s ucap=%.2f wss=%d' %
                (dist, dl, ucap, wss))

        # run test
        verbose('Running HRT test.')
        hrt_stats  = run_hrt_test(tp, int(wss), 'idle')
        hrt_stats += run_hrt_test(tp, int(wss), 'load')
        verbose('Running SRT test.')
        srt_stats  = run_srt_test(tp, int(wss), 'idle')
        srt_stats += run_srt_test(tp, int(wss), 'load')

        common = [dist, dl, options.machine, options.sched, wss, ucap]

        # produce output
        if options.header:
            print_header(out)
        util.print_rows([common + hrt_stats + srt_stats],
                        f=out, col_sep=', ', col_width=20,
                        prepend=' ')

        # close if not stdout
        if out != sys.stdout:
            out.close()
            del out
    else:
        print "Skipped; %s exists." % tname

def work_on_file(work):
    print work[0]
    try:
        process_file(work[1])
    except:
        print >>sys.stderr, "=" * 80
        print >>sys.stderr, "EXCEPTION:"
        traceback.print_exc(file=sys.stderr)
        print >>sys.stderr, "=" * 80
        raise
    sys.stdout.flush()

def process_files(files):
    fmt = "\n[%%0%dd/%d] %%s" % (len(str(len(files))), len(files))

    # randomize
    if options.randorder:
        random.shuffle(files)

    # parallelize
    if options.parallel:
        work = [(fmt % (i + 1, f), f) for (i, f) in enumerate(files)]
        p = multiprocessing.Pool()
        p.map(work_on_file, work)
    else:
        for i, f in enumerate(files):
            print fmt % (i + 1, f)
            process_file(f)
            sys.stdout.flush()

if __name__ == '__main__':
    parser = optparse.OptionParser(option_list=opts)
    parser.set_defaults(**defaults)
    (options, files) = parser.parse_args()

    if options.dump_header:
        print_header(sys.stdout)
        sys.exit(0)

    if not options.prefix is None:
        options.prefix = os.path.abspath(options.prefix)
        if not os.path.exists(options.prefix):
            print "Output dir doesn't exist: %s" % options.prefix
            exit(1)

    if options.glob:
        patterns = files
        files = []
        for p in patterns:
            expanded = os.path.expanduser(p)
            files.extend(glob(expanded))

    print 'sched=%s machine=%s quanta=%s VERSION=%s  cpp=%s' % \
        (options.sched, options.machine, options.quanta,
         VERSION, gedf.using_cpp)
    process_files(files)
