from math import ceil, floor

from overheads import Overheads, CPMD
from model.tasks import Task, TaskSystem
from binpack import Bin, MaxSpareCapacity

import gedf
import pfair

import util
from util import us2ms, ms2us
from util.quantor import forall

def load_overheads(host, scheduler, quanta=None, average=False):
    kind = 'avg' if average else 'wc'
    extra = '' if quanta is None else '_quanta=%s' % quanta
    fname = '%s/ohead-data/diss/oh_host=%s_scheduler=%s_stat=%s%s.csv' % \
        (util.get_this_repo(), host, scheduler, kind, extra)
    return Overheads.from_file(fname)

def load_lock_overheads(host, scheduler, locks, average=False):
    kind = 'avg' if average else 'wc'
    fname = '%s/ohead-data/diss/locks/oh_host=%s_scheduler=%s_locks=%s_stat=%s.csv' % \
        (util.get_this_repo(), host, scheduler, locks, kind)
    return Overheads.from_file(fname)

def load_cpmd(host, background, average=False):
    kind = 'avg' if average else 'wc'
    fname = '%s/ohead-data/diss/pmo_host=%s_background=%s_stat=%s.csv' % \
        (util.get_this_repo(), host, background, kind)
    return CPMD.from_file(fname)

def duplicate_taskset(taskset):
    new = []
    for t in taskset:
        new.append(Task(t.cost, t.period, t.deadline))
    return TaskSystem(new)

def duplicate_params(taskset):
    return [(t.cost, t.period, t.deadline) for t in taskset]

def taskset_from_params(params):
    return TaskSystem([Task(c, p, d) for (c, p, d) in params])

def charge_initial_load(taskset, oheads):
    if oheads:
        load = us2ms(oheads.initial_cache_load(taskset))
        if load > 0:
            for ti in taskset:
                ti.cost += load
                if ti.cost >= ti.period:
                    # infeasible
                    return False
    return True

def preemption_centric_irq_costs(taskset, oheads, dedicated_irq):
    qlen  = us2ms(oheads.quantum_size)
    tck   = us2ms(oheads.tick(taskset))
    ev_lat = us2ms(oheads.release_latency(taskset))

    # tick interrupt
    utick = tck / qlen

    urel  = 0.0
    if not dedicated_irq:
        rel   = us2ms(oheads.release(taskset))
        for ti in taskset:
            urel += (rel / ti.period)

    # cost of preemption
    cpre = tck + ev_lat * utick
    if not dedicated_irq:
        cpre += len(taskset) * rel + ev_lat * urel

    return (1.0 - utick - urel, cpre)

def account_for_oheads_edf(taskset, oheads, num_cpus, dedicated_irq):
    if not oheads:
        return True

    if not charge_initial_load(taskset, oheads):
        return False

    uscale, cpre = preemption_centric_irq_costs(taskset, oheads, dedicated_irq)

    if uscale <= 0:
        # interrupt overload
        return False

    sched = us2ms(2 * (oheads.sched(taskset)
                       + oheads.sched2(taskset)
                       + oheads.ctxsw(taskset))
                  + oheads.cache_affinity_loss(taskset))

    evlat = us2ms(oheads.release_latency(taskset))

    if dedicated_irq or num_cpus > 1:
        unscaled = 2 * cpre + us2ms(oheads.ipi_delay(taskset))
    else:
        unscaled = 2 * cpre

    for ti in taskset:
        ti.period   -= evlat
        ti.deadline -= evlat
        ti.cost      = (ti.cost + sched) / uscale + unscaled
        if ti.cost >= ti.period:
            return False

    return True

def hrt_test_edf(num_cpus, oheads, taskset,
                 dedicated_irq=False, total_cpus=None):

    if not account_for_oheads_edf(taskset, oheads, num_cpus,
                                  dedicated_irq):
        return False

    taskset.make_integral_us()
    return gedf.is_schedulable(num_cpus, taskset,
                               want_baker=False)

def srt_test_edf(num_cpus, oheads, taskset,
                 dedicated_irq=False, total_cpus=None):

    if not account_for_oheads_edf(taskset, oheads, num_cpus,
                                  dedicated_irq):
        return False

    taskset.make_integral_us()
    if gedf.is_schedulable(num_cpus, taskset,
                           want_baker=False):
        # great, zero tardiness
        bound = 0
    else:
        bound = gedf.soft.task_tardiness_bound(num_cpus, taskset,
                                               preemptive=True)
    if bound is None:
        return False
    else:
        for t in taskset:
            if bound > 0:
                t.completion = t.deadline + t.cost + bound
            else:
                t.completion = t.deadline
        return taskset

def stagger_latency(num_cpus, qlen):
    return (num_cpus - 1) / float(num_cpus) * qlen

def account_for_oheads_pd2(taskset,
                           oheads,
                           dedicated_irq,
                           staggered,
                           total_cpus):
    if not oheads:
        # Overhead-free execution; nothing to inflate.
        # Just make sure the parameters make sense.
        for ti in taskset:
            if ti.cost > ti.period:
                return False
        return True

    if not charge_initial_load(taskset, oheads):
        return False

    # account for release delay
    qlen = us2ms(oheads.quantum_size)

    ev_lat = us2ms(oheads.release_latency(taskset))
    rel_oh = us2ms(oheads.release(taskset))

    release_delay = qlen + ev_lat + rel_oh

    if staggered:
        release_delay += stagger_latency(total_cpus, qlen)

    # account for reduced effective quantum size
    qeff = qlen \
        - ev_lat \
        - us2ms(oheads.tick(taskset)) \
        - us2ms(oheads.sched(taskset)) \
        - us2ms(oheads.sched2(taskset)) \
        - us2ms(oheads.ctxsw(taskset)) \
        - us2ms(oheads.cache_affinity_loss(taskset))

    if not dedicated_irq:
        # account for release interrupts
        qeff -= (len(taskset) - 1) * rel_oh

    if qeff <= 0:
        return False

    # modify parameters of all tasks
    for ti in taskset:
        ti.cost   = qlen * ceil(ti.cost / qeff)
        ti.period = qlen * floor((ti.period - release_delay) / qlen)
        if ti.cost > ti.period:
            return False

    return True

# PD2 hard test. total_cpus is being used in the case of clustered PD2 to
# determine the correct stagger offset.
def hrt_test_pd2(num_cpus, oheads, taskset,
                 staggered=False, dedicated_irq=False, total_cpus=None):

    if not total_cpus:
        total_cpus = num_cpus

    if not account_for_oheads_pd2(taskset, oheads,
                                  dedicated_irq,
                                  staggered, total_cpus):
        return False
    else:
        return taskset.utilization() <= num_cpus

# PD2 soft test
def srt_test_pd2(num_cpus, oheads, taskset,
                 staggered=False, dedicated_irq=False, total_cpus=None):

    if not total_cpus:
        total_cpus = num_cpus

    if not account_for_oheads_pd2(taskset, oheads,
                                  dedicated_irq,
                                  False, total_cpus):
        return False
    elif taskset.utilization() <= num_cpus:
        if staggered and oheads:
            qlen = us2ms(oheads.quantum_size)
            tardiness = stagger_latency(total_cpus, qlen)
        else:
            tardiness = 0

        for t in taskset:
            # note: deadline is not reflected
            t.completion = t.period + ms2us(tardiness)

        return taskset
    else:
        # overutilized
        return False

def fp_inflate(oheads, taskset, dedicated_irq):
    T = taskset

    if not oheads:
        return T

    if not charge_initial_load(taskset, oheads):
        return False

    # jitter is given in integral microseconds!
    event_latency = int(ceil(oheads.release_latency(taskset)))

    tick_task = Task(us2ms(oheads.tick(T)), us2ms(oheads.quantum_size))
    tick_task.jitter = event_latency
    if tick_task.cost:
        tick_tasks = [tick_task]
    else:
        tick_tasks = []

    if not dedicated_irq:
        rel_cost   = us2ms(oheads.release(T))
        release_tasks = [Task(rel_cost, t.period) for t in taskset]
        for t in release_tasks:
            t.jitter = event_latency
    else:
        release_tasks = [] # don't impact us directly

    # account for initial release delay as jitter
    release_delay = event_latency + oheads.release(T)
    if dedicated_irq:
        release_delay += oheads.ipi_delay(T)
    release_delay = int(ceil(release_delay))

    for t in taskset:
        t.jitter = release_delay

    # account for scheduling cost and CPMD
    sched  = oheads.sched(T)
    sched2 = oheads.sched2(T)
    cxs    = oheads.ctxsw(T)
    cpmd   = oheads.cache_affinity_loss(T)
    preemption = us2ms(2 * (sched + sched2 + cxs) + cpmd)
    for t in taskset:
        t.cost += preemption

    return TaskSystem(tick_tasks + release_tasks + taskset)

def fp_demand(task, time):
    return task.cost * int(ceil(float(time + task.jitter) / task.period))

def fp_tda_schedulable(taskset, i):
    task = taskset[i]
    higher_prio = taskset[:i]

    # take into account that job is not immediately
    # availabe for scheduling
    # Equation (6) in ABRTW:93
    test_end = task.deadline - task.jitter

    time = sum([t.cost for t in higher_prio]) + task.cost
    while time <= test_end:
        demand = task.cost + sum([fp_demand(t, time) for t in higher_prio])
        if demand <= time:
            # yep, demand will be met by time
            task.completion = time
            return True
        else:
            # try again
            time = demand

    # if we get here, we didn't converge
    return False

def fp_tda_is_schedulable(taskset):
    assert forall(taskset)(lambda t: t.deadline <= t.period)
    for i, t in enumerate(taskset):
        if not fp_tda_schedulable(taskset, i):
            return False
    return True

# FP uniprocessor test
def hrt_test_fp(num_cpus, oheads, taskset,
                dedicated_irq=False, total_cpus=None):
    assert num_cpus == 1 # this is a uniprocessor test

    # assign deadline-monotonic priorities
    taskset.sort(key=lambda t: t.deadline)

    # This creates fake tasks to account for interrupts.
    taskset = fp_inflate(oheads, taskset, dedicated_irq)
    if not taskset:
        return False

    taskset.make_integral_us()
    return fp_tda_is_schedulable(taskset)

def srt_test_fp(num_cpus, oheads, taskset,
                dedicated_irq=False, total_cpus=None):
    if hrt_test_fp(num_cpus, oheads, taskset, dedicated_irq):
        return taskset
    else:
        return False

# partitioning heuristics
# need to take dedicated_irq handling into account

def partition_tasks(cluster_size, clusters, dedicated_irq, taskset):
    first_cap  = cluster_size - 1 if dedicated_irq else cluster_size
    first_bin  = Bin(size=Task.utilization, capacity=first_cap)
    other_bins = [Bin(size=Task.utilization, capacity=cluster_size) for _ in xrange(1, clusters)]

    heuristic = MaxSpareCapacity(initial_bins=[first_bin] + other_bins)

    # sort taskset in order of decreasing weights
    taskset.sort(key=Task.utilization, reverse=True)

    heuristic.binpack(taskset)

    if not heuristic.misfits:
        # Yep, they were placed alright.
        parts = [TaskSystem(b.items) for b in heuristic.bins]
        # keep track of CPU count
        for i, p in enumerate(parts):
            if i == 0 and dedicated_irq:
                p.cpus = cluster_size - 1
            else:
                p.cpus = cluster_size
        # Ignore bins without tasks.
        return [p for p in parts if len(p) > 0]
    else:
        # nope, didn't fit
        return False

def partitioned_test(sched_test, cluster_size, dedicated_irq, hard):

    # wrap given test in generic partitioning wrapper
    def part_test(num_cpus, oheads, taskset, *args, **kargs):
        assert num_cpus % cluster_size == 0
        partitions = partition_tasks(cluster_size, num_cpus / cluster_size,
                                     dedicated_irq, taskset)
        if not partitions:
            return False
        else:
            if hard:
                all = True
            else:
                all = []
            for p in partitions:
                res = sched_test(p.cpus, oheads, p, *args, **kargs)
                if not res:
                    # not schedulable
                    return False
                if not hard:
                    all.extend(res)
            # if we made it here, it's schedulable
            return all

    return part_test


def decode_scheduler_name(sched, cache_config, hrt):
    cluster_size = None
    dedicated_irq = False
    sched_test = None
    mem = 'MEM'

    # dedicated interrupt handling?
    if '-RM' in sched:
        dedicated_irq = True

    # basic policy?
    if 'EDF' in sched:
        sched_test = hrt_test_edf if hrt else srt_test_edf
    elif 'FP' in sched:
        sched_test = hrt_test_fp if hrt else srt_test_fp
    elif 'PD2' in sched:
        sched_test = hrt_test_pd2 if hrt else srt_test_pd2
    else:
        # Huh? Don't know this one.
        return None

    # partitioned or clustered scheduling?
    for k in cache_config:
        if k in sched:
            cluster_size = cache_config[k]
            mem = k
    if 'P-' in sched:
        cluster_size = 1
        mem = 'L1'

    if cluster_size:
        sched_test = partitioned_test(sched_test, cluster_size,
                                      dedicated_irq, hrt)
        is_global = False
    else:
        is_global = True

    return (sched_test, mem, dedicated_irq, is_global)


MACHINES = {
    'ludwig' : { 'L1' : 1, 'L2' : 2, 'L3' : 6, 'MEM' : 24}
}

def get_sched_test(host, sched, hrt=True):
    caches = MACHINES[host]
    return decode_scheduler_name(sched, caches, hrt)

def get_overheads(host, sched, background,
                  hrt, quanta=None, locks=None):
    oheads = load_overheads(host, sched, quanta, not hrt)

    if '-RM' in sched:
        # replace equivalent release latency from non-disturbed measurement
        alt = load_overheads(host, sched.replace('-RM', ''), quanta, not hrt)
        oheads.release_latency = alt.release_latency

    # set quantum size
    if 'PD2' in sched:
        # Pfair operates with HZ=1000
        oheads.quantum_size = ms2us(1)
    else:
        # event-driven schedulers work with HZ=100
        oheads.quantum_size = ms2us(10)

    # get CPMD
    cpmd = load_cpmd(host, background, not hrt)

    # check for locks
    if locks:
        loh = load_lock_overheads(host, sched, locks, not hrt)
        oheads.lock        = loh.lock
        oheads.unlock      = loh.unlock
        oheads.read_lock   = loh.read_lock
        oheads.read_unlock = loh.read_unlock

        # use syscall data from OMLP for spinlocks
        if locks in ['TF-Q', 'PF-Q', 'MX-Q']:
            loh = load_lock_overheads(host, sched,
                                      'FMLP' if 'FP' in sched else 'OMLP',
                                      not hrt)

        oheads.syscall_in  = loh.syscall_in
        oheads.syscall_out = loh.syscall_out


    return (oheads, cpmd)
