import ConfigParser
import sys
import os
import random
import copy
from itertools import izip
import multiprocessing as mp
from datetime import datetime
from collections import defaultdict

from functools import partial

import generator.tasksets as g
import util
from util import us2ms, storage, ms2us, atomic_create_file


from  model.tasks import TaskSystem
from overheads import CPMD, const, Overheads

import locking

import gedf
import fp

from exp.diss import get_overheads, fp_inflate
from exp.diss import partition_tasks, account_for_oheads_edf

MAX_NREQ = 5
CSLENGTH  = {
    'short'   : lambda: us2ms(random.uniform(1,   15)),
    'medium'  : lambda: us2ms(random.uniform(1,  100)),
    'long'    : lambda: us2ms(random.uniform(5, 1280)),
}

# config
# .pacc
# .wratio
# .nres
# .maxreq
# .length
#
# .wss
# .ucap
# .util
# .deadline
#
# .samples
# .fname


def draw_rw_count(config):
    reads  = 0
    writes = 0
    if random.random() <= config.pacc:
        # access resource
        nreqs = random.randint(1, config.maxreq)
        if config.wratio == 1 or random.random() <= config.wratio:
            writes = nreqs
        else:
            reads = nreqs
#        for _ in xrange(nreqs):
#            if config.wratio == 1 or random.random() <= config.wratio:
#                writes += 1
#            else:
#                reads  += 1
    return reads, writes

def generate_requests_for_task(config):
    requests = []
    for res_id in xrange(config.nres):
        num_reads, num_writes = draw_rw_count(config)
        length = CSLENGTH[config.cslen]
        if num_reads:
            requests.append( (res_id, num_reads, length(), locking.READ) )
        if num_writes:
            requests.append( (res_id, num_writes, length(), locking.WRITE) )
    return requests

def generate_requests(num_tasks, config):
    return [generate_requests_for_task(config) for _ in xrange(num_tasks)]

def ensure_write(config, treqs):
    written = [set() for _ in xrange(config.nres)]
    read    = [set() for _ in xrange(config.nres)]

    length = CSLENGTH[config.cslen]

    for i, reqs in enumerate(treqs):
        for (res_id, num, rlen, kind) in reqs:
            if kind == locking.READ:
                read[res_id].add(i)
            else:
                written[res_id].add(i)
    for i in xrange(config.nres):
        if len(written[i]) == 0 and len(read[i]) > 0:
            # need at least one writer
            x = random.randint(0, len(treqs) - 1)
            if x in read[i]:
                x = (x + 1) % len(treqs)
            treqs[x].append( (i, 1, length(), locking.WRITE) )
        elif len(written[i]) == 1 and len(read[i]) == 0:
            # need at least one reader
            x = random.randint(0, len(treqs) - 1)
            if x in written[i]:
                x = (x + 1) % len(treqs)
            treqs[x].append( (i, 1, length(), locking.READ) )


def generate_taskset(config):
    make_taskset = g.DIST[config.util]

    ts = make_taskset(config.ucap, None)
    reqs = generate_requests(len(ts), config)
#    if config.wratio < 1:
#        ensure_write(config, reqs)

    for (t, r) in izip(ts, reqs):
        t.resmodel = r

    return ts

def backup_params(ts):
    for t in ts:
        t.orig_cost      = t.cost
        t.orig_deadline  = t.deadline
        t.orig_period    = t.period
        t.orig_resmodel  = t.resmodel

def erase(t, key):
    if key in t.__dict__:
        del t.__dict__[key]

def restore_params(ts):
    ts.make_integral_us_called = False
    for t in ts:
        t.cost     = t.orig_cost
        t.deadline = t.orig_deadline
        t.period   = t.orig_period
        t.resmodel = t.orig_resmodel
        erase(t, 'blocked')
        erase(t, 'suspended')
        erase(t, 'arrival_blocked')
        erase(t, 'completion')
        erase(t, 'partition')
        erase(t, 'lock_prio')


SPINLOCKS = ['MX-Q', 'TF-Q', 'PF-Q']

spin_oheads = storage()

def get_oheads(host, sched, cache_measure, hard, locks=None, affinity='MEM'):
    (ohead, cpmd) = get_overheads(host, sched, cache_measure, hard, locks=locks)

    idx = CPMD.get_idx_for_name(affinity)
    ohead.cpmd_initial = cpmd.MEM
    ohead.cpmd_preempt = cpmd.mem_hierarchy[idx]

    return ohead

#    zero = Overheads()
#    zero.cpmd_initial = const(0)
#    zero.cpmd_preempt = const(0)
#    return zero


spin_oheads.pfp_hard_mtx = get_oheads('ludwig', 'P-FP-RM', 'load', True, locks='MX-Q', affinity='L1')
spin_oheads.pfp_hard_tf = get_oheads('ludwig', 'P-FP-RM', 'load', True, locks='TF-Q', affinity='L1')
spin_oheads.pfp_hard_pf = get_oheads('ludwig', 'P-FP-RM', 'load', True, locks='PF-Q', affinity='L1')

spin_oheads.pfp_soft_mtx = get_oheads('ludwig', 'P-FP-RM', 'idle', False, locks='MX-Q', affinity='L1')
spin_oheads.pfp_soft_tf = get_oheads('ludwig', 'P-FP-RM', 'idle', False, locks='TF-Q', affinity='L1')
spin_oheads.pfp_soft_pf = get_oheads('ludwig', 'P-FP-RM', 'idle', False, locks='PF-Q', affinity='L1')

spin_oheads.pedf_hard_mtx = spin_oheads.pfp_hard_mtx
spin_oheads.pedf_hard_tf = spin_oheads.pfp_hard_tf
spin_oheads.pedf_hard_pf = spin_oheads.pfp_hard_pf

spin_oheads.pedf_soft_mtx = spin_oheads.pfp_soft_mtx
spin_oheads.pedf_soft_tf = spin_oheads.pfp_soft_tf
spin_oheads.pedf_soft_pf = spin_oheads.pfp_soft_pf

spin_oheads.cedfl2_hard_mtx = get_oheads('ludwig', 'C-EDF-L2-RM', 'load', True, locks='MX-Q', affinity='L2')
spin_oheads.cedfl2_hard_tf  = get_oheads('ludwig', 'C-EDF-L2-RM', 'load', True, locks='TF-Q', affinity='L2')
spin_oheads.cedfl2_hard_pf  = get_oheads('ludwig', 'C-EDF-L2-RM', 'load', True, locks='PF-Q', affinity='L2')

spin_oheads.cedfl2_soft_mtx = get_oheads('ludwig', 'C-EDF-L2-RM', 'idle', False, locks='MX-Q', affinity='L2')
spin_oheads.cedfl2_soft_tf  = get_oheads('ludwig', 'C-EDF-L2-RM', 'idle', False, locks='TF-Q', affinity='L2')
spin_oheads.cedfl2_soft_pf  = get_oheads('ludwig', 'C-EDF-L2-RM', 'idle', False, locks='PF-Q', affinity='L2')

spin_oheads.cedfl3_hard_mtx = get_oheads('ludwig', 'C-EDF-L3-RM', 'load', True, locks='MX-Q', affinity='L3')
spin_oheads.cedfl3_hard_tf  = get_oheads('ludwig', 'C-EDF-L3-RM', 'load', True, locks='TF-Q', affinity='L3')
spin_oheads.cedfl3_hard_pf  = get_oheads('ludwig', 'C-EDF-L3-RM', 'load', True, locks='PF-Q', affinity='L3')

spin_oheads.cedfl3_soft_mtx = get_oheads('ludwig', 'C-EDF-L3-RM', 'idle', False, locks='MX-Q', affinity='L3')
spin_oheads.cedfl3_soft_tf  = get_oheads('ludwig', 'C-EDF-L3-RM', 'idle', False, locks='TF-Q', affinity='L3')
spin_oheads.cedfl3_soft_pf  = get_oheads('ludwig', 'C-EDF-L3-RM', 'idle', False, locks='PF-Q', affinity='L3')

spin_oheads.gedf_hard_mtx   = get_oheads('ludwig', 'G-EDF-RM', 'load', True, locks='MX-Q', affinity='MEM')
spin_oheads.gedf_hard_tf    = get_oheads('ludwig', 'G-EDF-RM', 'load', True, locks='TF-Q', affinity='MEM')
spin_oheads.gedf_hard_pf    = get_oheads('ludwig', 'G-EDF-RM', 'load', True, locks='PF-Q', affinity='MEM')

spin_oheads.gedf_soft_mtx   = get_oheads('ludwig', 'G-EDF-RM', 'idle', False, locks='MX-Q', affinity='MEM')
spin_oheads.gedf_soft_tf    = get_oheads('ludwig', 'G-EDF-RM', 'idle', False, locks='TF-Q', affinity='MEM')
spin_oheads.gedf_soft_pf    = get_oheads('ludwig', 'G-EDF-RM', 'idle', False, locks='PF-Q', affinity='MEM')

def spinlock_cluster_test(blocking_terms, sched_test, oheads,
                          total_cpus, cluster_size, config, taskset):

    # configure for correct WSS
    oheads.cache_affinity_loss = const(oheads.cpmd_preempt(config.wss))
    oheads.initial_cache_load  = const(oheads.cpmd_initial(config.wss))

    if total_cpus != cluster_size:
        partitions = partition_tasks(cluster_size, total_cpus / cluster_size, True, taskset)
        if not partitions:
            return False
    else:
        part = TaskSystem(taskset)
        part.cpus = total_cpus - 1
        partitions = [part]

    #  store where task is assigned
    for i, part in enumerate(partitions):
        cpu = i
        if cluster_size == 1:
            # dedicated interrupt handling: first partition is missing
            cpu += 1
        for t in part:
            t.partition = cpu

    # inflate locking-related overheads
    for part in partitions:
        if not locking.inflate_spin_locks(part, oheads):
            return False
        for t in part:
            # save for backup / restore
            t.inflated_resmodel = t.resmodel

    return sched_test(blocking_terms, oheads, cluster_size, taskset, partitions)


def charge_back_ipi_latency(tasks, oheads):
    ipilat = us2ms(oheads.ipi_delay(tasks))
    for t in tasks:
        # short np sections are masked by IPI latency
        reduction = min(ipilat, t.arrival_blocked)
        t.cost -= reduction

def hrt_test_edf_spin(blocking_terms, oheads, cluster_size, taskset, parts):
    for t in taskset:
        # assume task finishes by deadline <=> safe assumption under HRT
        t.completion = t.deadline

    # assign lock_prios according to relative deadlines
    locking.assign_edf_lock_prios(taskset)

    # bound blocking delays
    blocking_terms(taskset, cluster_size, 0)

 #   for t in taskset:
 #       print t, t.partition, t.lock_prio, t.blocked, t.arrival_blocked

    # extra charge sysin for arrival blocking (blocking job must call sched_yield())
    for part in parts:
        sysin = us2ms(oheads.syscall_in(part))
        for t in part:
            if t.arrival_blocked:
                # inflate for extra delay
                t.cost            += sysin
                t.arrival_blocked += sysin

    # account for scheduling overheads
    for (i, part) in enumerate(parts):
        charge_back_ipi_latency(part, oheads)
        if not account_for_oheads_edf(part, oheads, cluster_size, True):
#            print 'oheads failure'
            return False
        part.make_integral_us()
        if not gedf.is_schedulable(part.cpus, part,
                                   want_baker=False):
#            print 'sched failure'
            return False

    # great, everything passes
    return True


def srt_test_edf_spin(blocking_terms, oheads, cluster_size, taskset, parts):
    for t in taskset:
        # initially assume task finishes by deadline
        t.completion = t.deadline

    completion_ok = False

    count = 0
    while not completion_ok:
        completion_ok = True

        count += 1
        if count > 100:
            print 'SRT EDF SPIN GIVING UP: did not converge in 100 attempts'
            return False

        # assign lock_prios according to relative deadlines
        locking.assign_edf_lock_prios(taskset)

        # bound blocking delays with current assumption
        blocking_terms(taskset, cluster_size, 0)

        # extra charge sysin for arrival blocking (blocking job must call sched_yield())
        for part in parts:
            sysin = us2ms(oheads.syscall_in(part))
            for t in part:
                t.orig_partition = t.partition
                if t.arrival_blocked:
                    # inflate for extra delay
                    t.cost            += sysin
                    t.arrival_blocked += sysin

        # account for scheduling overheads
        for (i, part) in enumerate(parts):
            charge_back_ipi_latency(part, oheads)
            if not account_for_oheads_edf(part, oheads, part.cpus, True):
                return False
            part.make_integral_us()
            if  gedf.is_schedulable(part.cpus, part,
                                    want_baker=False):
                bound = 0
            else:
                bound = gedf.soft.task_tardiness_bound(part.cpus, part,
                                                       preemptive=True)
            if bound is None:
                # unbounded tardiness
                return False

            for t in part:
                if bound > 0:
                    t.new_completion = t.deadline + t.cost + bound
                else:
                    t.new_completion = t.deadline
                # completion was converted to us
                if t.new_completion > t.completion:
                    completion_ok = False # assumption invalidated


        if not completion_ok:
            # Ok, reiterate. Restore params from backup.
            for part in parts:
                restore_params(part)
                for t in part:
                    # reuse inflation
                    t.resmodel = t.inflated_resmodel
                    t.partition = t.orig_partition
                    # use new estimate
                    t.completion = us2ms(t.new_completion)

    # great, everything passes
    return True



def no_blocking_bounds(tasks, cluster_size, dedicated_irq):
    for t in tasks:
        t.suspended       = 0
        t.arrival_blocked = 0
        t.blocked         = 0
        t.local_blocking_count = 0


spin_tests = storage()

spin_tests.pedf_hard_none = partial(spinlock_cluster_test,
                                    no_blocking_bounds,
                                    hrt_test_edf_spin,
                                    spin_oheads.pedf_hard_mtx,
                                    24, 1)
spin_tests.pedf_hard_mtx = partial(spinlock_cluster_test,
                                   locking.apply_task_fair_mutex_bounds,
                                   hrt_test_edf_spin,
                                   spin_oheads.pedf_hard_mtx,
                                   24, 1)
spin_tests.pedf_hard_tf = partial(spinlock_cluster_test,
                                  locking.apply_task_fair_rw_bounds,
                                  hrt_test_edf_spin,
                                  spin_oheads.pedf_hard_tf,
                                  24, 1)
spin_tests.pedf_hard_pf = partial(spinlock_cluster_test,
                                  locking.apply_phase_fair_rw_bounds,
                                  hrt_test_edf_spin,
                                  spin_oheads.pedf_hard_pf,
                                  24, 1)
spin_tests.pedf_soft_none = partial(spinlock_cluster_test,
                                    no_blocking_bounds,
                                    srt_test_edf_spin,
                                    spin_oheads.pedf_soft_mtx,
                                    24, 1)
spin_tests.pedf_soft_mtx = partial(spinlock_cluster_test,
                                   locking.apply_task_fair_mutex_bounds,
                                   srt_test_edf_spin,
                                   spin_oheads.pedf_soft_mtx,
                                   24, 1)
spin_tests.pedf_soft_tf = partial(spinlock_cluster_test,
                                  locking.apply_task_fair_rw_bounds,
                                  srt_test_edf_spin,
                                  spin_oheads.pedf_soft_tf,
                                  24, 1)
spin_tests.pedf_soft_pf = partial(spinlock_cluster_test,
                                  locking.apply_phase_fair_rw_bounds,
                                  srt_test_edf_spin,
                                  spin_oheads.pedf_soft_pf,
                                  24, 1)


spin_tests.cedfl2_hard_none = partial(spinlock_cluster_test,
                                    no_blocking_bounds,
                                    hrt_test_edf_spin,
                                    spin_oheads.cedfl2_hard_mtx,
                                    24, 2)
spin_tests.cedfl2_hard_mtx = partial(spinlock_cluster_test,
                                   locking.apply_task_fair_mutex_bounds,
                                   hrt_test_edf_spin,
                                   spin_oheads.cedfl2_hard_mtx,
                                   24, 2)
spin_tests.cedfl2_hard_tf = partial(spinlock_cluster_test,
                                  locking.apply_task_fair_rw_bounds,
                                  hrt_test_edf_spin,
                                  spin_oheads.cedfl2_hard_tf,
                                  24, 2)
spin_tests.cedfl2_hard_pf = partial(spinlock_cluster_test,
                                  locking.apply_phase_fair_rw_bounds,
                                  hrt_test_edf_spin,
                                  spin_oheads.cedfl2_hard_pf,
                                  24, 2)
spin_tests.cedfl2_soft_none = partial(spinlock_cluster_test,
                                    no_blocking_bounds,
                                    srt_test_edf_spin,
                                    spin_oheads.cedfl2_soft_mtx,
                                    24, 2)
spin_tests.cedfl2_soft_mtx = partial(spinlock_cluster_test,
                                   locking.apply_task_fair_mutex_bounds,
                                   srt_test_edf_spin,
                                   spin_oheads.cedfl2_soft_mtx,
                                   24, 2)
spin_tests.cedfl2_soft_tf = partial(spinlock_cluster_test,
                                  locking.apply_task_fair_rw_bounds,
                                  srt_test_edf_spin,
                                  spin_oheads.cedfl2_soft_tf,
                                  24, 2)
spin_tests.cedfl2_soft_pf = partial(spinlock_cluster_test,
                                  locking.apply_phase_fair_rw_bounds,
                                  srt_test_edf_spin,
                                  spin_oheads.cedfl2_soft_pf,
                                  24, 2)


spin_tests.cedfl3_hard_none = partial(spinlock_cluster_test,
                                    no_blocking_bounds,
                                    hrt_test_edf_spin,
                                    spin_oheads.cedfl3_hard_mtx,
                                    24, 6)
spin_tests.cedfl3_hard_mtx = partial(spinlock_cluster_test,
                                   locking.apply_task_fair_mutex_bounds,
                                   hrt_test_edf_spin,
                                   spin_oheads.cedfl3_hard_mtx,
                                   24, 6)
spin_tests.cedfl3_hard_tf = partial(spinlock_cluster_test,
                                  locking.apply_task_fair_rw_bounds,
                                  hrt_test_edf_spin,
                                  spin_oheads.cedfl3_hard_tf,
                                  24, 6)
spin_tests.cedfl3_hard_pf = partial(spinlock_cluster_test,
                                  locking.apply_phase_fair_rw_bounds,
                                  hrt_test_edf_spin,
                                  spin_oheads.cedfl3_hard_pf,
                                  24, 6)
spin_tests.cedfl3_soft_none = partial(spinlock_cluster_test,
                                    no_blocking_bounds,
                                    srt_test_edf_spin,
                                    spin_oheads.cedfl3_soft_mtx,
                                    24, 6)
spin_tests.cedfl3_soft_mtx = partial(spinlock_cluster_test,
                                   locking.apply_task_fair_mutex_bounds,
                                   srt_test_edf_spin,
                                   spin_oheads.cedfl3_soft_mtx,
                                   24, 6)
spin_tests.cedfl3_soft_tf = partial(spinlock_cluster_test,
                                  locking.apply_task_fair_rw_bounds,
                                  srt_test_edf_spin,
                                  spin_oheads.cedfl3_soft_tf,
                                  24, 6)
spin_tests.cedfl3_soft_pf = partial(spinlock_cluster_test,
                                  locking.apply_phase_fair_rw_bounds,
                                  srt_test_edf_spin,
                                  spin_oheads.cedfl3_soft_pf,
                                  24, 6)


spin_tests.gedf_hard_none = partial(spinlock_cluster_test,
                                    no_blocking_bounds,
                                    hrt_test_edf_spin,
                                    spin_oheads.gedf_hard_mtx,
                                    24, 24)
spin_tests.gedf_hard_mtx = partial(spinlock_cluster_test,
                                   locking.apply_task_fair_mutex_bounds,
                                   hrt_test_edf_spin,
                                   spin_oheads.gedf_hard_mtx,
                                   24, 24)
spin_tests.gedf_hard_tf = partial(spinlock_cluster_test,
                                  locking.apply_task_fair_rw_bounds,
                                  hrt_test_edf_spin,
                                  spin_oheads.gedf_hard_tf,
                                  24, 24)
spin_tests.gedf_hard_pf = partial(spinlock_cluster_test,
                                  locking.apply_phase_fair_rw_bounds,
                                  hrt_test_edf_spin,
                                  spin_oheads.gedf_hard_pf,
                                  24, 24)
spin_tests.gedf_soft_none = partial(spinlock_cluster_test,
                                    no_blocking_bounds,
                                    srt_test_edf_spin,
                                    spin_oheads.gedf_soft_mtx,
                                    24, 24)
spin_tests.gedf_soft_mtx = partial(spinlock_cluster_test,
                                   locking.apply_task_fair_mutex_bounds,
                                   srt_test_edf_spin,
                                   spin_oheads.gedf_soft_mtx,
                                   24, 24)
spin_tests.gedf_soft_tf = partial(spinlock_cluster_test,
                                  locking.apply_task_fair_rw_bounds,
                                  srt_test_edf_spin,
                                  spin_oheads.gedf_soft_tf,
                                  24, 24)
spin_tests.gedf_soft_pf = partial(spinlock_cluster_test,
                                  locking.apply_phase_fair_rw_bounds,
                                  srt_test_edf_spin,
                                  spin_oheads.gedf_soft_pf,
                                  24, 24)

SPINLOCK_SETUPS = [
    ('P-EDF/None/HRT', spin_tests.pedf_hard_none),
    ('P-EDF/MX-Q/HRT', spin_tests.pedf_hard_mtx),
    ('P-EDF/TF-Q/HRT', spin_tests.pedf_hard_tf),
    ('P-EDF/PF-Q/HRT', spin_tests.pedf_hard_pf),

    ('P-EDF/None/SRT', spin_tests.pedf_soft_none),
    ('P-EDF/MX-Q/SRT', spin_tests.pedf_soft_mtx),
    ('P-EDF/TF-Q/SRT', spin_tests.pedf_soft_tf),
    ('P-EDF/PF-Q/SRT', spin_tests.pedf_soft_pf),

    ('C-EDF-L2/None/HRT', spin_tests.cedfl2_hard_none),
    ('C-EDF-L2/MX-Q/HRT', spin_tests.cedfl2_hard_mtx),
    ('C-EDF-L2/TF-Q/HRT', spin_tests.cedfl2_hard_tf),
    ('C-EDF-L2/PF-Q/HRT', spin_tests.cedfl2_hard_pf),

    ('C-EDF-L2/None/SRT', spin_tests.cedfl2_soft_none),
    ('C-EDF-L2/MX-Q/SRT', spin_tests.cedfl2_soft_mtx),
    ('C-EDF-L2/TF-Q/SRT', spin_tests.cedfl2_soft_tf),
    ('C-EDF-L2/PF-Q/SRT', spin_tests.cedfl2_soft_pf),


    ('C-EDF-L3/None/HRT', spin_tests.cedfl3_hard_none),
    ('C-EDF-L3/MX-Q/HRT', spin_tests.cedfl3_hard_mtx),
    ('C-EDF-L3/TF-Q/HRT', spin_tests.cedfl3_hard_tf),
    ('C-EDF-L3/PF-Q/HRT', spin_tests.cedfl3_hard_pf),

    ('C-EDF-L3/None/SRT', spin_tests.cedfl3_soft_none),
    ('C-EDF-L3/MX-Q/SRT', spin_tests.cedfl3_soft_mtx),
    ('C-EDF-L3/TF-Q/SRT', spin_tests.cedfl3_soft_tf),
    ('C-EDF-L3/PF-Q/SRT', spin_tests.cedfl3_soft_pf),

    ('G-EDF/None/HRT', spin_tests.gedf_hard_none),
    ('G-EDF/MX-Q/HRT', spin_tests.gedf_hard_mtx),
    ('G-EDF/TF-Q/HRT', spin_tests.gedf_hard_tf),
    ('G-EDF/PF-Q/HRT', spin_tests.gedf_hard_pf),

    ('G-EDF/None/SRT', spin_tests.gedf_soft_none),
    ('G-EDF/MX-Q/SRT', spin_tests.gedf_soft_mtx),
    ('G-EDF/TF-Q/SRT', spin_tests.gedf_soft_tf),
    ('G-EDF/PF-Q/SRT', spin_tests.gedf_soft_pf),


    ]

SPINLOCK_TESTS  = [x[1] for x in SPINLOCK_SETUPS]
SPINLOCK_TITLES = [x[0] for x in SPINLOCK_SETUPS]

def sample_spinlock(config):
    counts = [0] * len(SPINLOCK_TESTS)

    for _ in xrange(config.samples):
        ts = generate_taskset(config)
        backup_params(ts)
        for i, (name, is_schedulable) in enumerate(SPINLOCK_SETUPS):
#            print name, "*" * 80
            if is_schedulable(config, ts):
#                print 'schedulable'
                counts[i] += 1
            restore_params(ts)
    for i in xrange(len(counts)):
        counts[i] = counts[i] / float(config.samples)
    return [config] +  counts

def value_range(start, max, step):
    x  =  start
    while x <= max:
        yield x
        x += step

UCAP = list(value_range(1, 24.1, 0.5))
#UCAP = list(value_range(15, 15.1, 0.5))
#UCAP = [13]

def get_ucap_range(ref_config):
    configs = []
    for ucap in UCAP:
        config = copy.copy(ref_config)
        config.ucap = ucap
        configs.append(config)

    return configs

def spinlock_experiment_par(ref_config, out):

    configs = get_ucap_range(ref_config)

    pool = mp.Pool()
    data = pool.imap(sample_spinlock, configs, chunksize=1)

    util.print_row(['ucap'] + SPINLOCK_TITLES, f=out, col_width=20, prepend='#')
    out.flush()

    # write as we pull
    for row in data:
        row[0] = row[0].ucap
        util.print_row(row, f=out, col_width=20, col_sep=',', prepend=' ')
        out.flush()


def spinlock_experiment_seq(ref_config, out):

    configs = get_ucap_range(ref_config)

    util.print_row(['ucap'] + SPINLOCK_TITLES, f=out, col_width=20, prepend='#')
    out.flush()

    for conf in configs:
        row   = sample_spinlock(conf)
        row[0] = row[0].ucap
        util.print_row(row, f=out, col_width=20, col_sep=',', prepend=' ')
        out.flush()


spinlock_experiment = spinlock_experiment_par


def sob_cluster_test(blocking_terms, sched_test, oheads,
                     total_cpus, cluster_size, prio_inh, config, taskset):

    # configure for correct WSS
    oheads.cache_affinity_loss = const(oheads.cpmd_preempt(config.wss))
    oheads.initial_cache_load  = const(oheads.cpmd_initial(config.wss))

    if total_cpus != cluster_size:
        partitions = partition_tasks(cluster_size, total_cpus / cluster_size, True, taskset)
        if not partitions:
            return False
    else:
        part = TaskSystem(taskset)
        part.cpus = total_cpus - 1
        partitions = [part]

    #  store where task is assigned
    for i, part in enumerate(partitions):
        cpu = i
        if cluster_size == 1:
            # dedicated interrupt handling: first partition is missing
            cpu += 1
        for t in part:
            t.partition = cpu

    # inflate locking-related overheads
    for part in partitions:
        if not locking.inflate_semaphore_sob(part, oheads, prio_inh):
            return False
        for t in part:
            # save for backup / restore
            t.inflated_resmodel = t.resmodel

    return sched_test(blocking_terms, oheads, cluster_size, taskset, partitions,
                      prio_inh=prio_inh)



def hrt_test_edf_sob(blocking_terms, oheads, cluster_size, taskset, parts, prio_inh=False):
    for t in taskset:
        # assume task finishes by deadline <=> safe assumption under HRT
        t.completion = t.deadline

    # assign lock_prios according to relative deadlines
    if not prio_inh:
        locking.assign_edf_lock_prios(taskset)
    else:
        # not needed
        for t in taskset:
            t.lock_prio = 0

    # bound blocking delays
    blocking_terms(taskset, cluster_size, 0)

    if not prio_inh:
        # Clustered OMLP or partitioned OMLP.
        # Must charge extra scheduler invocation for one local
        # request that can block.
        for part in parts:
            switch_to = us2ms(oheads.sched(part) +
                              oheads.sched2(part) +
                              oheads.ctxsw(part))
            for t in part:
                if t.arrival_blocked:
                    # inflate for extra scheduler invocation
                    t.cost            += switch_to
                    t.arrival_blocked += switch_to

    # account for scheduling overheads
    for (i, part) in enumerate(parts):
        if not prio_inh:
            charge_back_ipi_latency(part, oheads)
        if not account_for_oheads_edf(part, oheads, cluster_size, True):
#            print 'oheads failure'
            return False
        part.make_integral_us()
        if not gedf.is_schedulable(part.cpus, part,
                                   want_baker=False):
#            print 'sched failure'
            return False
    return True


def srt_test_edf_sob(blocking_terms, oheads, cluster_size, taskset, parts,
                     prio_inh=False):
    for t in taskset:
        # initially assume task finishes by deadline
        t.completion = t.deadline

    completion_ok = False

    count = 0
    while not completion_ok:
        completion_ok = True

        count += 1
        if count > 100:
            print 'SRT EDF SOB GIVING UP: did not converge in 100 attempts'
            return False

        # assign lock_prios according to relative deadlines
        if not prio_inh:
            locking.assign_edf_lock_prios(taskset)
        else:
            # not needed
            for t in taskset:
                t.lock_prio = 0

        for t in taskset:
            t.orig_partition = t.partition

        # bound blocking delays with current assumption
        blocking_terms(taskset, cluster_size, 0)

        if not prio_inh:
            # Clustered OMLP or partitioned OMLP.
            # Must charge extra scheduler invocation for one local
            # request that can block.
            for part in parts:
                switch_to = us2ms(oheads.sched(part) +
                                  oheads.sched2(part) +
                                  oheads.ctxsw(part))
                for t in part:
                    if t.arrival_blocked:
                        # inflate for extra scheduler invocation
                        t.cost            += switch_to
                        t.arrival_blocked += switch_to

        # account for scheduling overheads
        for (i, part) in enumerate(parts):
            if not prio_inh:
                charge_back_ipi_latency(part, oheads)
            if not account_for_oheads_edf(part, oheads, part.cpus, True):
                return False
            part.make_integral_us()
            if  gedf.is_schedulable(part.cpus, part,
                                    want_baker=False):
                bound = 0
            else:
                bound = gedf.soft.task_tardiness_bound(part.cpus, part,
                                                       preemptive=True)
            if bound is None:
                # unbounded tardiness
                return False

            for t in part:
                if bound > 0:
                    t.new_completion = t.deadline + t.cost + bound
                else:
                    t.new_completion = t.deadline
                if t.new_completion > t.completion:
                    completion_ok = False # assumption invalidated


        if not completion_ok:
            # Ok, reiterate. Restore params from backup.
            for part in parts:
                restore_params(part)
                for t in part:
                    # reuse inflation
                    t.resmodel = t.inflated_resmodel
                    t.partition = t.orig_partition
                    # use new estimate
                    t.completion = us2ms(t.new_completion)

    return True


def fp_part_test(blocking_terms, oheads,
                 total_cpus, preemptive, config, taskset,
                 inflate_sem=locking.inflate_semaphore_saw):

    # configure for correct WSS
    oheads.cache_affinity_loss = const(oheads.cpmd_preempt(config.wss))
    oheads.initial_cache_load  = const(oheads.cpmd_initial(config.wss))

    partitions = partition_tasks(1, total_cpus, True, taskset)
    if not partitions:
        return False

    #  store where task is assigned
    for i, part in enumerate(partitions):
        # dedicated interrupt handling: first partition is missing
        cpu = i + 1
        for t in part:
            t.partition = cpu

    # inflate locking-related overheads
    for part in partitions:
        if not inflate_sem(part, oheads, preemptive):
            return False
        for t in part:
            # save for backup / restore
            t.inflated_resmodel = t.resmodel

    return test_fp_saw(blocking_terms, oheads, config, taskset, partitions,
                       preemptive=preemptive)


def dpcp_bounds(ts, config):
    locking.apply_dpcp_bounds(ts, 24, config.nres,
                              dedicated_irq=0)

def mpcp_vs_bounds(ts, config):
    locking.apply_mpcp_bounds(ts, True)

def mpcp_bounds(ts, config):
    locking.apply_mpcp_bounds(ts, False)

def fmlp_bounds(ts, config):
    locking.apply_part_fmlp_bounds(ts, True)

def fmlp_np_bounds(ts, config):
    locking.apply_part_fmlp_bounds(ts, False)

def test_fp_saw(blocking_terms, oheads, config, taskset, parts,
                preemptive=True):
    for t in taskset:
        # initially assume task finishes by execution time
        t.old_completion = int(ms2us(t.cost))

    # assign deadline-monotonic priorities
    taskset.sort(key=lambda t: t.deadline)
    for part in parts:
        part.sort(key=lambda t: t.deadline)

    completion_ok = False

    count = 0
    while not completion_ok:
        completion_ok = True

        count += 1
        if count > 100:
            print 'P-FP GIVING UP: did not converge in 100 attempts'
            return False

        locking.assign_fp_lock_prios(taskset)

        for t in taskset:
            t.completion = us2ms(t.old_completion)
#            print t, t.completion

        # bound blocking delays with current assumption
        blocking_terms(taskset, config)

        for t in taskset:
            t.orig_partition = t.partition

        if not preemptive:
            # Account for local blocking.
            # Add in one additional scheduler invocation.
            # This only affects the FMLP+.
            # The effect is that local blocking tasks actually delay    
            # even with the context switch after the lock was released.
            for part in parts:
                switch_to = us2ms(oheads.sched(part) +
                                  oheads.sched2(part) +
                                  oheads.ctxsw(part))
                for t in part:
                    if t.local_blocking_count:
                        # account for extra delay
                        t.blocked += switch_to * t.local_blocking_count

        # account for scheduling overheads
        for (i, part) in enumerate(parts):
            ts = fp_inflate(oheads, part, True)
            if not ts:
                return False
            ts.make_integral_us()

            for t in ts:
                # IRQ tasks are missing these
                if not 'suspended' in t.__dict__:
                    t.suspended = 0
                if not 'blocked' in t.__dict__:
                    t.blocked = 0
                if not 'extra_susp' in t.__dict__:
                    t.extra_susp = 0
                # convert to us, like the rest of the params
                t.suspended = int(ms2us(t.suspended))
                t.blocked   = int(ms2us(t.blocked))
                # extra suspension from overhead accounting
                x = int(ms2us(t.extra_susp))
                # jitter is already in us, see fp_inflate
                t.suspended += t.jitter + x
                t.blocked   += t.jitter + x
            if not fp.is_schedulable(ts):
                return False

            for t in part:
                if t.completion > t.old_completion:
                    completion_ok = False # assumption invalidated

        if not completion_ok:
            # Ok, reiterate. Restore params from backup.
            for part in parts:
                for t in part:
                    # use new estimate
                    t.old_completion = t.completion
                restore_params(part)
                for t in part:
                    # reuse inflation
                    t.resmodel = t.inflated_resmodel
                    t.partition = t.orig_partition

    return True



sem_oheads = storage()

sem_oheads.gedf_hard_fmlp  = get_oheads('ludwig', 'G-EDF-RM', 'load', True, locks='FMLP', affinity='MEM')
sem_oheads.gedf_hard_gomlp = get_oheads('ludwig', 'G-EDF-RM', 'load', True, locks='OMLP', affinity='MEM')
sem_oheads.gedf_hard_comlp = get_oheads('ludwig', 'G-EDF-RM', 'load', True, locks='C-OMLP', affinity='MEM')
sem_oheads.gedf_soft_fmlp  = get_oheads('ludwig', 'G-EDF-RM', 'idle', False, locks='FMLP', affinity='MEM')
sem_oheads.gedf_soft_gomlp = get_oheads('ludwig', 'G-EDF-RM', 'idle', False, locks='OMLP', affinity='MEM')
sem_oheads.gedf_soft_comlp = get_oheads('ludwig', 'G-EDF-RM', 'idle', False, locks='C-OMLP', affinity='MEM')

sem_oheads.cedfl3_hard_omlp = get_oheads('ludwig', 'C-EDF-L3-RM', 'load', True, locks='OMLP', affinity='L3')
sem_oheads.cedfl3_soft_omlp = get_oheads('ludwig', 'C-EDF-L3-RM', 'idle', False, locks='OMLP', affinity='L3')

sem_oheads.cedfl2_hard_omlp = get_oheads('ludwig', 'C-EDF-L2-RM', 'load', True, locks='OMLP', affinity='L2')
sem_oheads.cedfl2_soft_omlp = get_oheads('ludwig', 'C-EDF-L2-RM', 'idle', False, locks='OMLP', affinity='L2')

sem_oheads.pedf_hard_omlp = get_oheads('ludwig', 'C-EDF-L2-RM', 'load', True, locks='OMLP', affinity='L1')
sem_oheads.pedf_soft_omlp = get_oheads('ludwig', 'C-EDF-L2-RM', 'idle', False, locks='OMLP', affinity='L1')

sem_oheads.pfp_hard_fmlp = get_oheads('ludwig', 'P-FP-RM', 'load', True, locks='FMLP', affinity='L1')
sem_oheads.pfp_soft_fmlp = get_oheads('ludwig', 'P-FP-RM', 'idle', False, locks='FMLP', affinity='L1')

sem_oheads.pfp_hard_mpcp = get_oheads('ludwig', 'P-FP-RM', 'load', True, locks='MPCP', affinity='L1')
sem_oheads.pfp_soft_mpcp = get_oheads('ludwig', 'P-FP-RM', 'idle', False, locks='MPCP', affinity='L1')

sem_oheads.pfp_hard_mpcpvs = get_oheads('ludwig', 'P-FP-RM', 'load', True, locks='MPCP-VS', affinity='L1')
sem_oheads.pfp_soft_mpcpvs = get_oheads('ludwig', 'P-FP-RM', 'idle', False, locks='MPCP-VS', affinity='L1')

sem_oheads.pfp_hard_dpcp = get_oheads('ludwig', 'P-FP-RM', 'load', True, locks='DPCP', affinity='L1')
sem_oheads.pfp_soft_dpcp = get_oheads('ludwig', 'P-FP-RM', 'idle', False, locks='DPCP', affinity='L1')


sem_tests = storage()

sem_tests.pedf_hard_none = partial(sob_cluster_test,
                                   no_blocking_bounds,
                                   hrt_test_edf_sob,
                                   sem_oheads.pedf_hard_omlp,
                                   24, 1, False)
sem_tests.pedf_hard_omlp = partial(sob_cluster_test,
                                   locking.apply_clustered_omlp_bounds,
                                   hrt_test_edf_sob,
                                   sem_oheads.pedf_hard_omlp,
                                   24, 1, False)
sem_tests.cedfl2_hard_omlp = partial(sob_cluster_test,
                                     locking.apply_clustered_omlp_bounds,
                                     hrt_test_edf_sob,
                                     sem_oheads.cedfl2_hard_omlp,
                                     24, 2, False)
sem_tests.cedfl3_hard_omlp = partial(sob_cluster_test,
                                     locking.apply_clustered_omlp_bounds,
                                     hrt_test_edf_sob,
                                     sem_oheads.cedfl3_hard_omlp,
                                     24, 6, False)
sem_tests.gedf_hard_comlp = partial(sob_cluster_test,
                                    locking.apply_clustered_omlp_bounds,
                                    hrt_test_edf_sob,
                                    sem_oheads.gedf_hard_comlp,
                                    24, 24, False)
sem_tests.gedf_hard_gomlp = partial(sob_cluster_test,
                                    lambda ts, cpus, irq: locking.apply_global_omlp_bounds(ts, cpus),
                                    hrt_test_edf_sob,
                                    sem_oheads.gedf_hard_gomlp,
                                    24, 24, True)
sem_tests.gedf_hard_fmlp = partial(sob_cluster_test,
                                   lambda ts, cpus, irq: locking.apply_global_fmlp_sob_bounds(ts),
                                   hrt_test_edf_sob,
                                   sem_oheads.gedf_hard_fmlp,
                                   24, 24, True)

sem_tests.pedf_soft_none = partial(sob_cluster_test,
                                   no_blocking_bounds,
                                   srt_test_edf_sob,
                                   sem_oheads.pedf_soft_omlp,
                                   24, 1, False)
sem_tests.pedf_soft_omlp = partial(sob_cluster_test,
                                   locking.apply_clustered_omlp_bounds,
                                   srt_test_edf_sob,
                                   sem_oheads.pedf_soft_omlp,
                                   24, 1, False)
sem_tests.cedfl2_soft_omlp = partial(sob_cluster_test,
                                     locking.apply_clustered_omlp_bounds,
                                     srt_test_edf_sob,
                                     sem_oheads.cedfl2_soft_omlp,
                                     24, 2, False)
sem_tests.cedfl3_soft_omlp = partial(sob_cluster_test,
                                     locking.apply_clustered_omlp_bounds,
                                     srt_test_edf_sob,
                                     sem_oheads.cedfl3_soft_omlp,
                                     24, 6, False)
sem_tests.gedf_soft_comlp = partial(sob_cluster_test,
                                    locking.apply_clustered_omlp_bounds,
                                    srt_test_edf_sob,
                                    sem_oheads.gedf_soft_comlp,
                                    24, 24, False)
sem_tests.gedf_soft_gomlp = partial(sob_cluster_test,
                                    lambda ts, cpus, irq: locking.apply_global_omlp_bounds(ts, cpus),
                                    srt_test_edf_sob,
                                    sem_oheads.gedf_soft_gomlp,
                                    24, 24, True)
sem_tests.gedf_soft_fmlp = partial(sob_cluster_test,
                                   lambda ts, cpus, irq: locking.apply_global_fmlp_sob_bounds(ts),
                                   srt_test_edf_sob,
                                   sem_oheads.gedf_soft_fmlp,
                                   24, 24, True)


sem_tests.pfp_hard_fmlp = partial(fp_part_test,
                                  fmlp_bounds,
                                  sem_oheads.pfp_hard_fmlp,
                                  24, True)
sem_tests.pfp_hard_fmlp_np = partial(fp_part_test,
                                     fmlp_np_bounds,
                                     sem_oheads.pfp_hard_fmlp,
                                     24, False)
sem_tests.pfp_hard_mpcp = partial(fp_part_test,
                                  mpcp_bounds,
                                  sem_oheads.pfp_hard_mpcp,
                                  24, True)
sem_tests.pfp_hard_mpcpvs = partial(fp_part_test,
                                    mpcp_vs_bounds,
                                    sem_oheads.pfp_hard_mpcpvs,
                                    24, True)
sem_tests.pfp_hard_dpcp = partial(fp_part_test,
                                  dpcp_bounds,
                                  sem_oheads.pfp_hard_dpcp,
                                  24, True,
                                  inflate_sem=locking.inflate_dpcp)

sem_tests.pfp_soft_fmlp = partial(fp_part_test,
                                  fmlp_bounds,
                                  sem_oheads.pfp_soft_fmlp,
                                  24, True)
sem_tests.pfp_soft_fmlp_np = partial(fp_part_test,
                                     fmlp_np_bounds,
                                     sem_oheads.pfp_soft_fmlp,
                                     24, False)
sem_tests.pfp_soft_mpcp = partial(fp_part_test,
                                  mpcp_bounds,
                                  sem_oheads.pfp_soft_mpcp,
                                  24, True)
sem_tests.pfp_soft_mpcpvs = partial(fp_part_test,
                                    mpcp_vs_bounds,
                                    sem_oheads.pfp_soft_mpcpvs,
                                    24, True)
sem_tests.pfp_soft_dpcp = partial(fp_part_test,
                                  dpcp_bounds,
                                  sem_oheads.pfp_soft_dpcp,
                                  24, True,
                                  inflate_sem=locking.inflate_dpcp)


SEMAPHORE_SETUPS = [
    ('P-EDF/MX-Q/HRT', spin_tests.pedf_hard_mtx),
    ('C-EDF-L2/MX-Q/HRT', spin_tests.cedfl2_hard_mtx),
    ('C-EDF-L3/MX-Q/HRT', spin_tests.cedfl3_hard_mtx),
    ('G-EDF/MX-Q/HRT', spin_tests.gedf_hard_mtx),

    ('P-EDF/OMLP/HRT', sem_tests.pedf_hard_omlp),
    ('C-EDF-L2/OMLP/HRT', sem_tests.cedfl2_hard_omlp),
    ('C-EDF-L3/OMLP/HRT', sem_tests.cedfl3_hard_omlp),
    ('C-EDF-ALL/OMLP/HRT', sem_tests.gedf_hard_comlp),
    ('G-EDF/OMLP/HRT', sem_tests.gedf_hard_gomlp),
    ('G-EDF/FMLP/HRT', sem_tests.gedf_hard_fmlp),

    ('P-FP/FMLP/HRT', sem_tests.pfp_hard_fmlp),
    ('P-FP/FMLP-NP/HRT', sem_tests.pfp_hard_fmlp_np),
    ('P-FP/MPCP/HRT', sem_tests.pfp_hard_mpcp),
    ('P-FP/MPCP-VS/HRT', sem_tests.pfp_hard_mpcpvs),
    ('P-FP/DPCP/HRT', sem_tests.pfp_hard_dpcp),

    ('P-EDF/MX-Q/SRT', spin_tests.pedf_soft_mtx),
    ('C-EDF-L2/MX-Q/SRT', spin_tests.cedfl2_soft_mtx),
    ('C-EDF-L3/MX-Q/SRT', spin_tests.cedfl3_soft_mtx),
    ('G-EDF/MX-Q/SRT', spin_tests.gedf_soft_mtx),

    ('P-EDF/OMLP/SRT', sem_tests.pedf_soft_omlp),
    ('C-EDF-L2/OMLP/SRT', sem_tests.cedfl2_soft_omlp),
    ('C-EDF-L3/OMLP/SRT', sem_tests.cedfl3_soft_omlp),
    ('C-EDF-ALL/OMLP/SRT', sem_tests.gedf_soft_comlp),
    ('G-EDF/OMLP/SRT', sem_tests.gedf_soft_gomlp),
    ('G-EDF/FMLP/SRT', sem_tests.gedf_soft_fmlp),

    ('P-FP/FMLP/SRT', sem_tests.pfp_soft_fmlp),
    ('P-FP/FMLP-NP/SRT', sem_tests.pfp_soft_fmlp_np),
    ('P-FP/MPCP/SRT', sem_tests.pfp_soft_mpcp),
    ('P-FP/MPCP-VS/SRT', sem_tests.pfp_soft_mpcpvs),
    ('P-FP/DPCP/SRT', sem_tests.pfp_soft_dpcp),

    ]

SEMAPHORE_TITLES = [x[0] for x in SEMAPHORE_SETUPS]

def sample_semaphore(config):
    counts = [0] * len(SEMAPHORE_SETUPS)

    for _ in xrange(config.samples):
        ts = generate_taskset(config)
        backup_params(ts)
        for i, (name, is_schedulable) in enumerate(SEMAPHORE_SETUPS):
#            print name, "*" * 80
            if is_schedulable(config, ts):
#                print 'schedulable'
                counts[i] += 1
            restore_params(ts)
    for i in xrange(len(counts)):
        counts[i] = counts[i] / float(config.samples)
    return [config] +  counts


def semaphore_experiment_par(ref_config, out):

    configs = get_ucap_range(ref_config)

    pool = mp.Pool()
    data = pool.imap(sample_semaphore, configs, chunksize=1)

    util.print_row(['ucap'] + SEMAPHORE_TITLES, f=out, col_width=20, prepend='#')
    out.flush()

    # write as we pull
    for ucap, row in izip(UCAP, data):
        assert ucap == row[0].ucap
        row[0] = row[0].ucap
        if sum(row[1:]) == 0:
            break
        util.print_row(row, f=out, col_width=20, col_sep=',', prepend=' ')
        out.flush()

    last_ucap = ucap
    # see if we need to fill in zeros
    if last_ucap < UCAP[-1]:
        row = [0] + [0.0 for _ in SEMAPHORE_TITLES]
        for ucap in UCAP:
            if ucap >= last_ucap:
                row[0] = ucap
                util.print_row(row, f=out, col_width=20, col_sep=',', prepend=' ')
        out.flush()


def semaphore_experiment_seq(ref_config, out):

    configs = get_ucap_range(ref_config)

    util.print_row(['ucap'] + SEMAPHORE_TITLES, f=out, col_width=20, prepend='#')
    out.flush()

    for conf in configs:
        row   = sample_semaphore(conf)
        row[0] = row[0].ucap
        util.print_row(row, f=out, col_width=20, col_sep=',', prepend=' ')
        out.flush()

semaphore_experiment = semaphore_experiment_par

def run_config(config):
    output = atomic_create_file(config.fname)
    if output is None:
        print "Skipping '%s', file exists." % config.fname
        return

    start  = datetime.now()
    output.write('# Start: %s\n' % datetime.now())


    head     = util.git_current_head()
    modified = util.git_locally_modified()
    output.write("# git HEAD: %s%s\n" % (head if head else "<unkown>",
                                         '+' if modified else ''))

    for k in config:
        output.write('# %s=%s\n' % (k, config[k]))

    if config.comp == 'spinlocks':
        spinlock_experiment(config, output)
    elif config.comp == 'semaphores':
        semaphore_experiment(config, output)

    end = datetime.now()
    output.write('# Duration: %s\n' % (end - start))
    output.close()

def load_config(fname):
    cfg = ConfigParser.RawConfigParser()
    cfg.read(fname)

    config = storage()

    config.util      = cfg.get('tasks', 'util')
    config.wss       = cfg.getfloat('tasks', 'wss')
    config.deadlines = cfg.get('tasks', 'deadlines')

    config.pacc   = cfg.getfloat('locks', 'preq')
    config.wratio = cfg.getfloat('locks', 'wratio')
    config.nres   = cfg.getint('locks', 'nres')
    config.maxreq = cfg.getint('locks', 'maxreq')
    config.cslen  = cfg.get('locks', 'cslen')

    config.samples = cfg.getint('experiment', 'samples')
    config.fname   = cfg.get('experiment', 'fname')
    config.comp    = cfg.get('experiment', 'comparison')

#    for x in config.keys():
#        print x, '->', config[x]

    config.config_name = fname

    return config

def die(msg):
    print >> sys.stderr, msg
    sys.exit(1)

def main(args=sys.argv[1:]):
    if not args:
        die("usage: diss_lock.py <config file>")
    else:
        for i, arg in enumerate(reversed(args)):
            if os.path.exists('stop-lock-exp'):
                print 'Aborting as requested...'
                break
            try:
                config = load_config(arg)
            except ConfigParser.Error as msg:
                die("%s: %s" % (args[0], msg))
            print "[%d/%d] %s %s -> %s" % (i+ 1, len(args), datetime.now(),
                                           arg, config.fname)
            run_config(config)
        print 'Terminating.'

if __name__ == '__main__':
    main()
