from math import ceil
from collections import defaultdict

import cpptest as cpp
from util import us2ms, ms2us

READ  = 'r'
WRITE = 'w'

def convert_param(x):
    return int(ceil(ms2us(x)))

def convert_blocking(x):
    return us2ms(x)

# assumes mutex constraints
def get_cpp_model(all_tasks, convert=convert_param, use_task_period=False):
    rsi = cpp.ResourceSharingInfo(len(all_tasks))
    for t in all_tasks:
        rsi.add_task(
            convert(t.period),
            convert(t.period) if use_task_period else convert(t.completion),
            t.partition,
            t.lock_prio)
        # use defaultdict since some resource may be both read and written
        reqs  = defaultdict(int)
        lens  = defaultdict(int)
        for req in t.resmodel:
            id  = req[0]
            reqs[id] += req[1]
            lens[id] = max(req[2], lens[id])

        for id in reqs:
            num    = reqs[id]
            length = lens[id]
            rsi.add_request_rw(id, num, convert(length), cpp.WRITE)
    return rsi

def get_cpp_model_rw(all_tasks, convert=convert_param):
    rsi = cpp.ResourceSharingInfo(len(all_tasks))
    for t in all_tasks:
        rsi.add_task(convert(t.period), convert(t.completion), t.partition, t.lock_prio)
        for req in t.resmodel:
            # presume mutual exclusion unless specified otherwise
            kind = cpp.WRITE
            id  = req[0]
            num = req[1]
            length = req[2]
            if len(req) == 4 and req[3] == READ:
                kind = cpp.READ
            rsi.add_request_rw(id, num, convert(length), kind)
    return rsi

def assign_edf_lock_prios(all_tasks):
    all_deadlines = set([t.deadline for t in all_tasks])
    prio = {}
    for i, dl in enumerate(all_deadlines):
        prio[int(dl)] = i

    for t in all_tasks:
        t.lock_prio = prio[int(t.deadline)]

def assign_fp_lock_prios(all_tasks):
    # prioritized in index order
    for i, t in enumerate(all_tasks):
        t.lock_prio = i

def apply_mpcp_bounds(all_tasks, use_virtual_spin=False):
    model = get_cpp_model(all_tasks)
    res = cpp.mpcp_bounds(model, use_virtual_spin)

    if use_virtual_spin:
        for i,t in enumerate(all_tasks):
            # no suspension time
            t.suspended = 0
            # all blocking, including arrival blocking
            t.blocked   = convert_blocking(res.get_blocking_term(i))
            # remote blocking increases CPU demand (s-oblivious)
            t.cost += convert_blocking(res.get_remote_blocking(i))
    else:
        for i,t in enumerate(all_tasks):
            # remote blocking <=> suspension time
            t.suspended = convert_blocking(res.get_remote_blocking(i))
            # all blocking, including arrival blocking
            t.blocked   = convert_blocking(res.get_blocking_term(i))


# default resource assignment: round robin
def apply_dpcp_bounds(all_tasks, num_cpus, num_resources,
                      dedicated_irq=cpp.NO_CPU):
    # The DPCP bounds are expressed in terms of task periods,
    # not response time.
    model = get_cpp_model(all_tasks, use_task_period=True)
    loc   = cpp.ResourceLocality()

    for res_id in xrange(num_resources):
        cpu = res_id % num_cpus
        if cpu == dedicated_irq:
            cpu = (cpu + 1) % num_cpus
        loc.assign_resource(res_id, cpu)
    res = cpp.dpcp_bounds(model, loc)

    for i,t in enumerate(all_tasks):
        # remote blocking <=> suspension time
        t.suspended = convert_blocking(res.get_remote_blocking(i))
        # all blocking, including arrival blocking
        t.blocked   = convert_blocking(res.get_blocking_term(i))


def apply_part_fmlp_bounds(all_tasks, preemptive=True):
    model = get_cpp_model(all_tasks)
    res = cpp.part_fmlp_bounds(model, preemptive)

    for i,t in enumerate(all_tasks):
        # remote blocking <=> suspension time
        t.suspended = convert_blocking(res.get_remote_blocking(i))
        # all blocking, including local blocking
        t.blocked   = convert_blocking(res.get_blocking_term(i))
        t.local_blocking_count = res.get_local_count(i)

def apply_suspension_oblivious(all_tasks, res):
    for i,t in enumerate(all_tasks):
        # s-oblivious <=> no suspension
        t.suspended = 0
        # might be zero
        t.arrival_blocked = convert_blocking(res.get_arrival_blocking(i))
        # all blocking, including arrival blocking
        t.blocked   = convert_blocking(res.get_blocking_term(i))
        # s-oblivious: charge it as execution cost
        t.cost     += t.blocked

# S-oblivious bounds only.
def apply_global_fmlp_sob_bounds(all_tasks):
    model = get_cpp_model(all_tasks)
    res = cpp.global_fmlp_bounds(model)
    apply_suspension_oblivious(all_tasks, res)

def apply_global_omlp_bounds(all_tasks, num_cpus):
    model = get_cpp_model(all_tasks)
    res = cpp.global_omlp_bounds(model, num_cpus)
    apply_suspension_oblivious(all_tasks, res)

def apply_clustered_omlp_bounds(all_tasks, procs_per_cluster,
                                dedicated_irq=cpp.NO_CPU):
    model = get_cpp_model(all_tasks)
    res = cpp.clustered_omlp_bounds(model, procs_per_cluster, dedicated_irq)
    apply_suspension_oblivious(all_tasks, res)

def apply_clustered_rw_omlp_bounds(all_tasks, procs_per_cluster,
                                   dedicated_irq=cpp.NO_CPU):
    model = get_cpp_model_rw(all_tasks)
    res = cpp.clustered_rw_omlp_bounds(model, procs_per_cluster, dedicated_irq)
    apply_suspension_oblivious(all_tasks, res)

# spinlocks are charged similarly to s-oblivious analysis

def apply_task_fair_mutex_bounds(all_tasks, procs_per_cluster,
                                 dedicated_irq=cpp.NO_CPU):
    model = get_cpp_model(all_tasks)
    res = cpp.task_fair_mutex_bounds(model, procs_per_cluster, dedicated_irq)
    apply_suspension_oblivious(all_tasks, res)

def apply_task_fair_rw_bounds(all_tasks, procs_per_cluster,
                              dedicated_irq=cpp.NO_CPU):
    model = get_cpp_model_rw(all_tasks)
    # mutex equivalent model
    model_mtx = get_cpp_model(all_tasks)
    res = cpp.task_fair_rw_bounds(model, model_mtx, procs_per_cluster, dedicated_irq)
    apply_suspension_oblivious(all_tasks, res)

def apply_phase_fair_rw_bounds(all_tasks, procs_per_cluster,
                                 dedicated_irq=cpp.NO_CPU):
    model = get_cpp_model_rw(all_tasks)
    res = cpp.phase_fair_rw_bounds(model, procs_per_cluster, dedicated_irq)
    apply_suspension_oblivious(all_tasks, res)



# Overhead accounting.

# This assumes absence of any interrupts.
def inflate_spin_locks(tasks, oheads):
    # the individual charges
    rcost  = us2ms(oheads.read_lock(tasks) + oheads.read_unlock(tasks))
    wcost  = us2ms(oheads.lock(tasks) + oheads.unlock(tasks))
    sysin  = us2ms(oheads.syscall_in(tasks))
    sysout = us2ms(oheads.syscall_out(tasks))

    # inflate each request and each task's exec cost
    for t in tasks:
        infmodel   = []
        extra_costs = 0.0

        for req in t.resmodel:
            # presume mutual exclusion unless specified otherwise
            kind = WRITE
            id     = req[0]
            num    = req[1]
            length = req[2]
            if len(req) == 4 and req[3] == READ:
                kind = READ
                inf = length + rcost
            else:
                inf = length + wcost
            infmodel.append((id, num, inf, kind))
            extra_costs += (inf + sysin + sysout) * num

        t.cost    += extra_costs
        t.resmodel = infmodel
        if t.cost >= t.deadline:
            return False

    return True


def inflate_semaphore_saw(tasks, oheads, preemptive):
    return inflate_semaphore(tasks, oheads, preemptive, True)

def inflate_semaphore_sob(tasks, oheads, preemptive):
    return inflate_semaphore(tasks, oheads, preemptive, False)

# This assumes absence of any interrupts.
def inflate_semaphore(tasks, oheads, preemptive, suspension_aware):
    lock   = us2ms(oheads.lock(tasks))
    unlock = us2ms(oheads.unlock(tasks))
    sysin  = us2ms(oheads.syscall_in(tasks))
    sysout = us2ms(oheads.syscall_out(tasks))
    sched  = us2ms(oheads.sched(tasks) +
                   oheads.sched2(tasks) +
                   oheads.ctxsw(tasks))
    cpmd  = us2ms(oheads.cache_affinity_loss(tasks))
    ipi   = us2ms(oheads.ipi_delay(tasks))

    # per-request execution cost increase (equ 7.3)
    # 3 sched: wait + resume + yield
    exec_increase = 3 * sched + \
        2 * sysin  + \
        2 * sysout + \
        1 *   lock + \
        1 * unlock + \
        2 * cpmd

    # delay to be boken up
    if suspension_aware:
        susp_increase = ipi
    else:
        # s-oblivious: count IPI as execution time
        susp_increase = 0
        exec_increase += ipi

    # For non-preemptive protocols, this is the remote case.
    # Additional local costs are charged separately.
    # This only affects the FMLP+, the partitioned OMLP, and the
    # clustered OMLP.
    cs_increase = ipi + sched + sysout + sysin + unlock

    # preemptive protocols, add in additional scheduling cost
    if preemptive:
        cs_increase += sched

    # inflate each request and each task's exec cost
    for t in tasks:
        infmodel   = []
        xtra_cost  = 0
        count = 0

        for req in t.resmodel:
            # currently only mutex
            id     = req[0]
            num    = req[1]
            length = req[2]
            inf = length + cs_increase
            xtra_cost += length * num
            count += num
            infmodel.append( (id, num, inf) )

        t.cost        += xtra_cost + count * exec_increase
        t.extra_susp   = count * susp_increase
        t.resmodel = infmodel
        if t.cost >= t.deadline:
            return False

    return True


def inflate_dpcp(tasks, oheads, preemptive):
    lock   = us2ms(oheads.lock(tasks))
    unlock = us2ms(oheads.unlock(tasks))
    sysin  = us2ms(oheads.syscall_in(tasks))
    sysout = us2ms(oheads.syscall_out(tasks))
    sched  = us2ms(oheads.sched(tasks) +
                   oheads.sched2(tasks) +
                   oheads.ctxsw(tasks))
    cpmd  = us2ms(oheads.cache_affinity_loss(tasks))
    ipi   = us2ms(oheads.ipi_delay(tasks))

    exec_increase = 2 * sched + sysin  + sysout + lock + unlock + 2 * cpmd
    cs_increase   = 3 * sched + sysin + sysout + lock + unlock
    susp_increase = 2 * ipi + cs_increase

    # inflate each request and each task's exec cost
    for t in tasks:
        infmodel   = []
        xtra_cost  = 0
        xtra_susp  = 0

        for req in t.resmodel:
            # currently only mutex
            id     = req[0]
            num    = req[1]
            length = req[2]
            inf = length + cs_increase
            xtra_cost += exec_increase * num
            xtra_susp += (susp_increase + length) * num
            infmodel.append( (id, num, inf) )

        t.cost        += xtra_cost
        t.extra_susp   = xtra_susp
        t.resmodel = infmodel

        if t.cost > t.deadline or t.extra_susp > t.deadline:
            return False

    return True
