import math
import csv
import types

from util import print_rows, print_row, load_column_csv, storage

# all overheads in microseconds

def delistify(x):
    if isinstance(x, types.ListType):
        x = len(x)
    return float(x)

class LinearEqu(object):
    def __init__(self, a, b, coerce=lambda x: x):
        self.a = a
        self.b = b
        self.coerce = coerce

    def __str__(self):
        slope = round(self.b, 3)
        if abs(slope) >= 0.001:
            return '%.3f%+.3f n' % (self.a, slope)
        else:
            return '%.3f' % self.a

    def __call__(self, x):
        return self.a + (self.b * self.coerce(x) if self.b else 0)

    def __add__(self, other):
        return LinearEqu(self.a + other.a, self.b + other.b, self.coerce)

    def __mul__(self, scalar):
        return LinearEqu(self.a * scalar, self.b * scalar, self.coerce)

    def __rmul__(self, scalar):
        return self * scalar

    def is_constant(self):
        return self.b == 0

class PieceWiseLinearEqu(object):
    def __init__(self, points, coerce=lambda x: x,
                 force_monotonicly_increasing=False):
        # points = [(x1, y1), (x2, y2), ...]
        assert len(points) >= 2

        def slope(i):
            dy = points[i+1][1] - points[i][1]
            dx = points[i+1][0] - points[i][0]
            return dy / dx

        def yintercept(i):
            x, y = points[i]
            dy = slope(i) * x
            return y - dy

        self.coerce   = coerce
        self.segments = [(points[i+1][0],
                          LinearEqu(yintercept(i), slope(i),
                                    lambda x: x))
                         for i in xrange(len(points) - 1)]

    def __call__(self, x):
        # find piece (slow and stupid)
        x = self.coerce(x)
        for (limit, f) in self.segments:
            if limit > x:
                break
        # approximate linearly from support point
        # negative overheads make no sense, so avoid them
        return max(0, f(x))

    def is_constant(self):
        return all([seg[1].is_constant() for seg in self.segments])

def const(x):
    return LinearEqu(x, 0)

def lin(a, b):
    return LinearEqu(a, b, coerce=delistify)

def scale(alpha, fun):
    return lambda x: fun(x) * alpha

def piece_wise_linear(points, coerce=delistify):
    return PieceWiseLinearEqu(points, coerce)

def make_monotonic(points):
    filtered = points[:1]
    prevprev = None
    _, prev = filtered[0]
    for (x, y) in points[1:]:
        # y values should not decrease
        y = max(prev, y)
        if not prevprev is None and prevprev == y:
            # remove useless intermediate point
            filtered.pop()
        filtered.append((x,y))
        prevprev, prev = prev, y

    # also remove the last one if it is not needed (i.e., constant)
    if len(filtered) == 2 and filtered[-1][1] == filtered[-2][1]:
        filtered.pop()

    return filtered

def is_monotonic(points):
    x1, y1 = points[0]
    for (x2, y2) in points[1:]:
        assert x1 < x2
        if y1 > y2:
            return False
        x1, y1 = x2, y2
    return True

def monotonic_pwlin(points, coerce=delistify):
    ascending = make_monotonic(points)
    if len(ascending) > 1:
        assert is_monotonic(ascending)
        return piece_wise_linear(ascending, coerce)
    elif ascending:
        return const(ascending[0][1])
    else:
        return const(0)


class Overheads(object):
    """Legacy overhead objects"""
    def __init__(self):
        self.zero_overheads()

    FIELD_MAPPING = [
        ('SEND_RESCHED', 'ipi_delay'),
        ('SCHED',        'sched'),
        ('SCHED2',       'sched2'),
        ('RELEASE',      'release'),
        ('CXS',          'ctxsw'),
        ('TICK',         'tick'),
        ('PMO',          'cache_affinity_loss'),
        ('PULL_TIMER',   'pull_timer'),
        ('RELEASE_LATENCY', 'release_latency'),
        ('LOCK',         'lock'),
        ('UNLOCK',       'unlock'),
        ('READ_LOCK',    'read_lock'),
        ('READ_UNLOCK',  'read_unlock'),
        ('SYSCALL_IN',   'syscall_in'),
        ('SYSCALL_OUT',  'syscall_out'),
        ]

    def zero_overheads(self):
        self.quantum_size = 1000 # microseconds

        self.cache_affinity_loss = const(0) # cache-related preemption/migration delay

        for (name, field) in self.FIELD_MAPPING:
            self.__dict__[field] = const(0)

    def zero_non_irq_overheads(self):
        for (name, field) in self.FIELD_MAPPING:
            if not field in ['release', 'ipi_delay', 'tick']:
                self.__dict__[field] = const(0)

    def __str__(self):
        return " ".join(["%s: %s" % (name, self.__dict__[field])
                         for (name, field) in Overheads.FIELD_MAPPING])

    def scale(self, factor):
        for (name, field) in self.FIELD_MAPPING:
            self.__dict__[field] = scale(factor, self.__dict__[field])

    def get_profile(self, points):
        return [[x] + [self.__dict__[field](x)
                 for (_, field) in Overheads.FIELD_MAPPING]
                for x in points]

    def write_profile(self, fname, points, want_header=True,
                      comment_header=False):
        f = open(fname, 'w')
        if want_header:
            print_row(['NUM_TASKS'] +
                      [name for (name,_) in Overheads.FIELD_MAPPING],
                      col_sep=' ,',
                      f=f,
                      prepend='#' if comment_header else ' ',
                      col_width=20)
        table = self.get_profile(points)
        print_rows(table, f=f, col_sep=',', prepend=' ', col_width=20)
        f.close()

    @staticmethod
    def from_file(fname, non_decreasing=True):
        # no error checking---crash loudly if something is amiss
        f = open(fname)
        d = list(csv.reader(f))
        f.close()
        col_idx = {}
        for i, key in enumerate(d[0]):
            col_idx[key.strip()] = i
        data    = d[1:]

        def col(i):
            for row in data:
                if row:
                    yield float(row[i])

        if not 'NUM_TASKS' in col_idx:
            raise IOError, "no NUM_TASKS?"

        o = Overheads()
        o.source = fname

        for (name, field) in Overheads.FIELD_MAPPING:
            if name in col_idx:
                points = zip(col(col_idx['NUM_TASKS']), col(col_idx[name]))
                if non_decreasing:
                    o.__dict__[field] = monotonic_pwlin(points)
                else:
                    o.__dict__[field] = piece_wise_linear(points)
        return o


class SystemOverheads(object):
    """New-style overhead object"""
    def __init__(self, reference=Overheads(), n=0,
                 migration_cost=None, const_reference=True):
        if const_reference:
            self.tick    = const(reference.tick(n))
            self.sched   = const(reference.sched(n) + reference.sched2(n))
            self.release = const(reference.release(n))
            self.timer_migration = const(reference.pull_timer(n))
            self.ipi_latency = const(reference.ipi_delay(n))
            self.context_switch = const(reference.ctxsw(n))
        else:
            self.tick    = reference.tick
            self.sched   = lambda x: (reference.sched(x) + reference.sched2(x))
            self.release = reference.release
            self.timer_migration = reference.pull_timer
            self.ipi_latency = reference.ipi_delay
            self.context_switch = reference.ctxsw
        if migration_cost:
            self.cpmd = migration_cost
        else:
            self.cpmd = lambda x, y: 0.0


class CPMD(object):

    MEM, L1, L2, L3 = 0, 1, 2, 3
    MAPPING = list(enumerate(["MEM", "L1", "L2", "L3"]))

    def __init__(self, l1=0, l2=0, l3=0, mem=0):
        self.mem_hierarchy  = [const(mem), const(l1), const(l2), const(l3)]
        for (i, name) in CPMD.MAPPING:
            self.__dict__[name] = self.mem_hierarchy[i]

    def cpmd_cost(self, shared_mem_level, working_set_size):
        return self.mem_hierarch[shared_mem_level](working_set_size)

    def max_cost(self, working_set_size):
        return max([f(working_set_size) for f in self.mem_hierarchy])

    @staticmethod
    def get_idx_for_name(key):
        for (i, name) in CPMD.MAPPING:
            if name == key:
                return i
        assert False # bad key

    @staticmethod
    def from_file(fname, non_decreasing=True):
        id = lambda x: x
        data = load_column_csv(fname, convert=float)
        if not 'WSS' in data.by_name:
            raise IOError, 'no WSS?'

        o = CPMD()

        for idx, name in CPMD.MAPPING:
            if name in data.by_name:
                points = zip(data.by_name['WSS'], data.by_name[name])
                if non_decreasing:
                    o.mem_hierarchy[idx] = monotonic_pwlin(points, coerce=id)
                else:
                    o.mem_hierarchy[idx] = piece_wise_linear(points, coerce=id)
                o.__dict__[name] = o.mem_hierarchy[idx]
        return o

    def get_profile(self, points):
        return [[x] + [cost(x)
                       for cost in self.mem_hierarchy]
                for x in points]

    def write_profile(self, fname, points, want_header=True,
                      comment_header=False):
        f = open(fname, 'w')
        if want_header:
            print_row(['WSS'] +
                      [name for (_,name) in CPMD.MAPPING],
                      col_sep=' ,',
                      f=f,
                      prepend='#' if comment_header else ' ',
                      col_width=20)
        table = self.get_profile(points)
        print_rows(table, f=f, col_sep=',', prepend=' ', col_width=20)
        f.close()


def write_interpolated_files(files):
    for file in files:
        print "Generating profile for %s..." % file
        if 'pmo' in file:
            points = range(0, 10000, 50)
            o = CPMD.from_file(file)
        else:
            points = range(1, 500)
            o = Overheads.from_file(file)
        o.write_profile('overhead-profile-' + os.path.basename(file), points)

def write_interpolated_builtin():
    profiles = [(Overheads.gedf_soft, "gedf-soft"),
                (Overheads.gedf_hard, "gedf-hard"),
                (Overheads.gsnedf_soft, "gsnedf-soft"),
                (Overheads.gsnedf_hard, "gsnedf-hard"),
                ]
    points = range(1, 500)
    for f, name in profiles:
        print "Generating profile for %s..." % name
        o = Overheads()
        f(o)
        o.write_profile('overhead-profile-' + name + '.csv', points)


if __name__ == '__main__':
    import sys
    import os.path

    if len(sys.argv) == 2 and sys.argv[1] == 'builtin':
        write_interpolated_builtin()
    else:
        write_interpolated_files(sys.argv[1:])
