from util      import ms2us, us2ms
from util.quantor   import forall
from quanta    import has_integral_period, quantize_wcet
from math import ceil, floor

def account_for_overheads(overheads, tasks, release_master=False,
                          periodic=True,
                          initial_cache_cost_implicit=True):
    """We consider the quantum size to be an
    overhead (since ideally we would have a quantum size of 0).

    The effective quantum size is the amount of time that a job
    can use each quantum to make progress:

    effective quantum size = quantum size - tick - max(migration, preempt)
    - sched - ctxsw

    The inflated job execution cost is then found by the number of
    quanta required to satisfy the uninflated execution cost.

    Account for overheads after everything else has been taken into account.
    """
    T = tasks
    qsize = overheads.quantum_size

    # make sure that all periods are multiple of the quantum size
    assert forall(T)(has_integral_period(us2ms(qsize)))

    equant  = float(qsize)
    equant -= overheads.tick(T)
    equant -= overheads.sched(T)
    equant -= overheads.sched2(T)
    equant -= overheads.ctxsw(T)
    equant -= overheads.cache_affinity_loss(T)
    if not periodic and not release_master:
        # account for release interrupts
        equant -= overheads.release(T) * (len(T) - 1)
    if equant <= 0:
        return False
    else:
        if not initial_cache_cost_implicit:
            ccost = us2ms(overheads.cache_affinity_loss(T))
            for t in T:
                t.cost += ccost
        quantize_wcet(us2ms(qsize), T, us2ms(equant))
        return T

def is_schedulable(no_cpus, tasks):
    return tasks.utilization() <= no_cpus and \
        forall(tasks)(lambda t: t.period >= t.cost)


def pseudo_release(subtask, util):
    return int(floor((subtask - 1) / util))

def pseudo_deadline(subtask, util):
    return int(ceil(subtask / util))

def boundary_bit(subtask, util):
    return int(ceil(subtask / util) - floor(subtask / util))

def group_deadline(subtask, util):
    if 2 * util >= 1:
        return int(ceil(((pseudo_deadline(subtask, util) - subtask)
                         / (1 - util) )))
    else:
        return 0

def subtasks(tsk, quantum_size=1):
    util   = tsk.utilization_q()
    quanta = int(ceil(tsk.cost / float(quantum_size)))
    return [(pseudo_release(k, util),
             pseudo_deadline(k, util),
             boundary_bit(k, util),
             group_deadline(k, util)) for k in xrange(1, quanta + 1)]

